﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Toolkit.Algorithm {
	/// <summary>
	/// 遺伝的アルゴリズム
	/// </summary>
	public class GA {
		/// <summary>
		/// 遺伝子
		/// </summary>
		public interface IGene : ICloneable {
			/// <summary>
			/// 突然変異
			/// </summary>
			void Mutate();
		}

		/// <summary>
		/// 具体的な処理の実装
		/// </summary>
		public interface IStrategy {
			/// <summary>
			/// 適応度の評価
			/// </summary>
			/// <param name="gene">遺伝子</param>
			/// <returns>適応度</returns>
			int Evaluate(IGene gene);
			/// <summary>
			/// 交差。
			/// c1とc2を交差させた結果をc1とc2へ出力。
			/// 事前にIGene.Clone()してあるので(その実装がちゃんとしてれば)そのまま書き換えてもOK。
			/// </summary>
			void Crossing(ref IGene c1, ref IGene c2);
		}

		static Random rand = new Random();

		/// <summary>
		/// 処理の実装
		/// </summary>
		IStrategy strategy;

		/// <summary>
		/// 遺伝子リスト
		/// </summary>
		List<Entry> list = new List<Entry>();
		/// <summary>
		/// 遺伝子データ
		/// </summary>
		class Entry {
			public IGene Gene;
			public int? Fitness;
		}

		/// <summary>
		/// 突然変異確率
		/// </summary>
		public double MutateProbability { get; set; }

		/// <summary>
		/// その時点で最も優秀な遺伝子
		/// </summary>
		public IGene Top {
			get { return list[0].Gene; }
		}

		/// <summary>
		/// 初期化
		/// </summary>
		public GA(IStrategy strategy) {
			this.strategy = strategy;
		}

		/// <summary>
		/// 遺伝子の追加
		/// </summary>
		/// <param name="gene">遺伝子</param>
		public void Add(IGene gene) {
			Entry e = new Entry();
			e.Gene = gene;
			list.Add(e);
		}

		/*
			// 交叉と淘汰の処理
			generation.MakeGenes(townCount);
			for (int i = 0; i < generationCount; i++) {
				generation.ToNext(townCount, valueDatas);
			}
		 */

		/// <summary>
		/// 遺伝子の生成
		/// </summary>
		/*
		public void MakeGenes(int geneCount, int townCount) {
			for (int i = 0; i < geneCount; i++) {
				list.Add(new GAGene(townCount));
			}
		}
		*/

		/// <summary>
		/// 世代交代
		/// </summary>
		public void DoStep() {
			// 交叉
			int index1 = rand.Next(0, list.Count);
			int index2;
			do {
				index2 = rand.Next(0, list.Count);
			} while (index1 == rand.Next(0, list.Count));
			IGene g1 = (IGene)list[index1].Gene.Clone();
			IGene g2 = (IGene)list[index2].Gene.Clone();
			strategy.Crossing(ref g1, ref g2);

			// 淘汰
			Selection(g1, g2);

			// 突然変異
			foreach (var gene in list) {
				if (rand.NextDouble() <= MutateProbability) {
					gene.Gene.Mutate();
				}
			}
		}

		/// <summary>
		/// 淘汰
		/// </summary>
		public void Selection(IGene g1, IGene g2) {
			list.Add(new Entry { Gene = g1, Fitness = strategy.Evaluate(g1) });
			list.Add(new Entry { Gene = g2, Fitness = strategy.Evaluate(g2) });

			// 適応度の降順にソート
			list.Sort((x, y) => y.Fitness.Value.CompareTo(x.Fitness.Value));

			// 適応度の低い遺伝子をリストから削除
			list.RemoveAt(list.Count - 1);
			list.RemoveAt(list.Count - 1);
		}

	}
}
