﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;
using System.Threading;

namespace ふせん {
	public partial class DummyForm : Form {
		Config config = new Config();
		Data data = new Data();
		List<TagForm> tagForms = new List<TagForm>();
		TagDrawer drawer;

		bool topMost = false;

		/// <summary>
		/// タグファイル保存場所。あちこちで使うしコレの移動はサポートしないのでメンバに持ってしまう。
		/// </summary>
		readonly string tagsDir;

		public DummyForm() {
			InitializeComponent();
			Visible = false;
			Rectangle screenRect = new Rectangle();
			foreach (Screen s in Screen.AllScreens) { screenRect = Rectangle.Union(screenRect, s.Bounds); }
			Location = new Point(screenRect.Left - Size.Width, screenRect.Top - Size.Height);

			// コンフィグの読み込み。
			data = Data.Deserialize();
			config = Config.Deserialize();

			drawer = new TagDrawer(config);

			// tagsDirの設定
			string dir = Path.GetDirectoryName(Process.GetCurrentProcess().MainModule.FileName);
			tagsDir = Path.Combine(dir, "tags");

			int build = System.Diagnostics.Process.GetCurrentProcess()
				.MainModule.FileVersionInfo.FileBuildPart;
			notifyIcon1.Text = "ふせん build-" + build;
			notifyIcon1.Visible = true;
		}

		/// <summary>
		/// 非アクティブに表示
		/// </summary>
		protected override bool ShowWithoutActivation {
			get { return true; }
		}

		/// <summary>
		/// 初期化。
		/// </summary>
		private void DummyForm_Load(object sender, EventArgs e) {
			Visible = false;

			// ディレクトリが存在しなければ作成
			if (!Directory.Exists(tagsDir)) {
				Directory.CreateDirectory(tagsDir);
			}

			lock (tagForms) {
				// ディレクトリの監視開始
				fileSystemWatcher1.Path = tagsDir;
				fileSystemWatcher1.NotifyFilter = NotifyFilters.FileName | NotifyFilters.LastWrite | NotifyFilters.Size;
				fileSystemWatcher1.EnableRaisingEvents = true;

				// ディレクトリ内のファイルを片っ端から付箋化
				foreach (string file in Directory.GetFiles(tagsDir, "*.txt", SearchOption.TopDirectoryOnly)) {
					string tagFileName = GetTagFileName(file);
					if (tagFileName == null) {
						// 謎ファイル。。
						System.Diagnostics.Debug.Fail("謎なファイル名：" + file);
					} else {
						TagData d = data.TagDataList.Find(delegate(TagData t) {
							return t.FileName.ToLower() == tagFileName.ToLower();
						});
						if (d == null) {
							// 既存のが無いっぽいので新たに作成
							d = new TagData(tagFileName);
							data.TagDataList.Add(d);
						}
						// 付箋を表示。
						CreateTagForm(d);
					}
				}

				// ついでに使われてないTagDataを削除する
				GarbageTagDataList();

				// 一応保存しとく
				data.WindowHandle = Handle.ToInt64(); // ついでにハンドルも。
				data.Serialize();
			}
		}
		private void DummyForm_Shown(object sender, EventArgs e) {
			Visible = false;
		}

		/// <summary>
		/// 後始末。
		/// </summary>
		private void DummyForm_FormClosing(object sender, FormClosingEventArgs e) {
			notifyIcon1.Dispose();
			// 付箋たちを削除。
			foreach (TagForm form in tagForms) {
				form.Dispose();
			}
			// ↑でLocationがdataに書き込まれたはずなので保存
			data.WindowHandle = 0; // ついでにハンドルも。
			data.Serialize();
		}

		/// <summary>
		/// TagForm追加処理
		/// </summary>
		private TagForm CreateTagForm(TagData d) {
			TagForm form = new TagForm(d, tagsDir);
			form.OnEdit += new TagForm.EditDelegate(tagForm_OnEdit);
			form.Show(this);
			form.ReloadTagFile(config, drawer, true);
			form.SetTopMost(topMost);
			foreach (TagForm f in tagForms) {
				f.OnNewForm(form);
			}
			tagForms.Add(form);
			data.Serialize();
			return form;
		}

		/// <summary>
		/// TagForm削除処理
		/// </summary>
		private void DeleteTagForm(TagForm form) {
			foreach (TagForm f in tagForms) {
				f.OnDeleteForm(form);
			}
			tagForms.Remove(form);
			form.Dispose(); // 後始末も忘れずに。。
			data.Serialize();
		}

		/// <summary>
		/// data.TagDataListの要らないデータを削除っとく。
		/// </summary>
		private void GarbageTagDataList() {
			List<TagData> tagDataList = new List<TagData>();
			foreach (TagForm form in tagForms) {
				tagDataList.Add(form.TagData);
			}
			data.TagDataList = tagDataList;
		}

		/// <summary>
		/// タグファイル名（tagsDirからの相対パス）を取得
		/// </summary>
		private string GetTagFileName(string file) {
			if (file.Length <= tagsDir.Length ||
				file.Substring(0, tagsDir.Length) != tagsDir ||
				file[tagsDir.Length] != Path.DirectorySeparatorChar) {
				// 謎ファイル。。
				System.Diagnostics.Debug.Fail("謎なファイル名：" + file);
				return null;
			} else {
				return file.Substring(tagsDir.Length + 1);
			}
		}

		/// <summary>
		/// 編集するときの通知
		/// </summary>
		void tagForm_OnEdit(TagForm sender, EventArgs e) {
			sender.Edit(config);
		}

		#region メニューな処理

		private void 付箋の追加AToolStripMenuItem_Click(object sender, EventArgs e) {
			for (int i = 0; ; i++) {
				string file = string.Format("{0}\\{1}.txt", tagsDir, i.ToString("d3"));
				if (!File.Exists(file)) {
					// 空ファイルを作成するだけ。あとはFileSystemWatcher側でちゃんと処理られるはず。。
					File.WriteAllBytes(file, new byte[] { 0xef, 0xbb, 0xbf });
					break;
				}
			}
		}

		private void 最前面表示TToolStripMenuItem_Click(object sender, EventArgs e) {
			topMost = !topMost;
			最前面表示TToolStripMenuItem.Checked = topMost;
			foreach (TagForm form in tagForms) {
				form.SetTopMost(topMost);
			}
		}

		private void 一時的に隠すSToolStripMenuItem_Click(object sender, EventArgs e) {
			一時的に隠すSToolStripMenuItem.Enabled = false;
			再表示TToolStripMenuItem.Enabled = true;
			foreach (TagForm f in tagForms) f.Hide();
		}

		private void 再表示TToolStripMenuItem_Click(object sender, EventArgs e) {
			foreach (TagForm f in tagForms) f.Show();
			再表示TToolStripMenuItem.Enabled = false;
			一時的に隠すSToolStripMenuItem.Enabled = true;
		}

		private void 設定CToolStripMenuItem_Click(object sender, EventArgs e) {
			using (ConfigForm form = new ConfigForm(config)) {
				if (form.ShowDialog(this) == DialogResult.OK) {
					config = form.Config;
					config.Serialize();
					// 設定の反映
					lock (tagForms) {
						drawer = new TagDrawer(config);
						foreach (TagForm f in tagForms) {
							f.ReloadTagFile(config, drawer, true);
						}
					}
				}
			}
		}

		private void 実行ファイルのあるフォルダを開くMToolStripMenuItem_Click(object sender, EventArgs e) {
			System.Diagnostics.Process.Start(System.IO.Path.GetDirectoryName(Application.ExecutablePath));
		}

		private void 配布サイトを開くWToolStripMenuItem_Click(object sender, EventArgs e) {
			System.Diagnostics.Process.Start("http://hp.vector.co.jp/authors/VA039571/");
		}

		private void 作者へメールを送るEToolStripMenuItem_Click(object sender, EventArgs e) {
			System.Diagnostics.Process.Start("http://tqzh.tk:24497/site/mail/");
		}

		private void ネットワーク更新設定OToolStripMenuItem_Click(object sender, EventArgs e) {
			using (Toolkit.Updater.UpdateConfigForm form = new Toolkit.Updater.UpdateConfigForm(config.UpdateConfig)) {
				if (form.ShowDialog(this) == DialogResult.OK) {
					config.UpdateConfig = form.Config;
					config.Serialize();
				}
			}
		}

		private void ネットワーク更新NToolStripMenuItem_Click(object sender, EventArgs e) {
			//string url = "http://tqzh.tk:24497/data/fs.cfg";
			string url = "http://hp.vector.co.jp/authors/VA039571/fs.cfg";
			using (Toolkit.Updater.UpdateForm form = new Toolkit.Updater.UpdateForm(url,
				config.UpdateConfig, data.UpdateRecord)) {
				form.ShowDialog(this);
				data.UpdateRecord = form.UpdateRecord;
				data.Serialize();
				if (form.Finished) {
					Close();
				}
			}
		}

		private void 再起動RToolStripMenuItem_Click(object sender, EventArgs e) {
			Toolkit.App.App.SetRestart();
			Close();
		}

		private void 終了XToolStripMenuItem_Click(object sender, EventArgs e) {
			Close();
		}

		#endregion

		#region fileSystemWatcherな処理

		/// <summary>
		/// ファイルサイズとか変わった風味
		/// </summary>
		private void fileSystemWatcher1_Changed(object sender, FileSystemEventArgs e) {
			lock (tagForms) {
				string tagFileName = GetTagFileName(Toolkit.IO.Utility.GetCorrectPath(e.FullPath));
				Debug.Assert(tagFileName != null);
				if (tagFileName == null) return;

				try {
					InnerReloadTagFile(tagFileName);
				} catch (Exception ex) {
					Debug.Fail(ex.ToString());
					ThreadPool.QueueUserWorkItem(arg => { 
						while (true) {
							Thread.Sleep(500); // 適当に寝つつ。

							Exception exc = null;
							Invoke(new MethodInvoker(delegate() {
								try {
									InnerReloadTagFile(tagFileName);
								} catch (Exception exce) {
									exc = exce;
								}
							}));
							if (exc == null) {
								break;
							} else {
								Debug.Fail(exc.ToString());
							}
						}
					});
				}
			}
		}

		private void InnerReloadTagFile(string tagFileName) {
			TagForm form = tagForms.Find(delegate(TagForm t) { return t.TagData.FileName.ToLower() == tagFileName.ToLower(); });
			if (form != null) {
				form.ReloadTagFile(config, drawer, false);
			}
		}

		/// <summary>
		/// ファイル作成された
		/// </summary>
		private void fileSystemWatcher1_Created(object sender, FileSystemEventArgs e) {
			lock (tagForms) {
				string tagFileName = GetTagFileName(Toolkit.IO.Utility.GetCorrectPath(e.FullPath));
				Debug.Assert(tagFileName != null);
				if (tagFileName == null) return;

				TagForm form = CreateTagForm(new TagData(tagFileName));
				data.TagDataList.Add(form.TagData);

				// 編集
				form.Edit(config);
			}
		}

		/// <summary>
		/// ファイル削除された
		/// </summary>
		private void fileSystemWatcher1_Deleted(object sender, FileSystemEventArgs e) {
			lock (tagForms) {
				string tagFileName = GetTagFileName(Toolkit.IO.Utility.GetCorrectPath(e.FullPath));
				Debug.Assert(tagFileName != null);
				if (tagFileName == null) return;

				TagForm form = tagForms.Find(delegate(TagForm t) { return t.TagData.FileName.ToLower() == tagFileName.ToLower(); });
				if (form != null) {
					data.TagDataList.Remove(form.TagData);
					DeleteTagForm(form);
				}
			}
		}

		/// <summary>
		/// ファイルがリネームされた
		/// </summary>
		private void fileSystemWatcher1_Renamed(object sender, RenamedEventArgs e) {
			lock (tagForms) {
				string tagFileName = GetTagFileName(e.OldFullPath);
				Debug.Assert(tagFileName != null);
				if (tagFileName == null) return;

				TagForm form = tagForms.Find(delegate(TagForm t) { return t.TagData.FileName.ToLower() == tagFileName.ToLower(); });
				if (form != null) {
					form.OnChangeFileName(GetTagFileName(Toolkit.IO.Utility.GetCorrectPath(e.FullPath)));
					data.Serialize();
				}
			}
		}

		#endregion

		/// <summary>
		/// タスクトレイアイコンがダブルクリックされた
		/// </summary>
		private void notifyIcon1_DoubleClick(object sender, EventArgs e) {
			switch (config.DoubleClickAction) {
			case 0: 付箋の追加AToolStripMenuItem_Click(this, null); break;
			case 1: 最前面表示TToolStripMenuItem_Click(this, null); break;
			case 2: 設定CToolStripMenuItem_Click(this, null); break;
			}
		}
	}
}