﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ふせん {
	/// <summary>
	/// タグの描画を行うクラス
	/// </summary>
	public class TagDrawer : IDisposable {
		Config config;
		Font font;
		Brush fontBrush;
		Brush decoBrush;
		Brush bgBrush;

		public Font Font { get { return font; } }

		public TagDrawer(Config config) {
			this.config = config;

			font = new Font(config.FontFamily, config.FontSize, config.FontStyle);
			fontBrush = new SolidBrush((Color)config.FontColor);
			decoBrush = new SolidBrush((Color)config.DecorationColor);
			bgBrush = new SolidBrush((Color)config.TagColor);
		}

		public void Dispose() {
			bgBrush.Dispose();
			decoBrush.Dispose();
			fontBrush.Dispose();
			font.Dispose();
		}

		/// <summary>
		/// サイズの取得
		/// </summary>
		public Size GetSize(Graphics g, string text) {
			using (StringFormat sf = new StringFormat()) {
				sf.FormatFlags = StringFormatFlags.FitBlackBox | StringFormatFlags.NoClip | StringFormatFlags.NoWrap;
				if (config.AntiAlias) {
					g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
				} else {
					g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.SingleBitPerPixelGridFit;
				}
				SizeF size = g.MeasureString(text, font, int.MaxValue, sf);
				Size s = new Size((int)(size.Width * g.PageScale), (int)(size.Height * g.PageScale));
				s.Width += config.TagMargin.Width * 2;
				s.Height += config.TagMargin.Height * 2;
				if (config.DecorationType == 1) { // 影
					int gap = GetShadowGap();
					s.Width += gap;
					s.Height += gap;
				} else if (config.DecorationType == 2) { // 縁取り
					int gap = GetHemGap();
					s.Width += gap * 2;
					s.Height += gap * 2;
				}
				if (s.Width < 8) s.Width = 8;
				if (s.Height < 8) s.Height = 8;
				return s;
			}
		}

		/// <summary>
		/// 描画
		/// </summary>
		public void Draw(Graphics g, string text, Rectangle clientRectangle) {
			g.FillRectangle(bgBrush, clientRectangle);
			using (StringFormat sf = new StringFormat()) {
				sf.FormatFlags = StringFormatFlags.FitBlackBox | StringFormatFlags.NoClip | StringFormatFlags.NoWrap;
				if (config.AntiAlias) {
					g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
				} else {
					g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.SingleBitPerPixelGridFit;
				}

				// マージンの分描画領域を縮める
				clientRectangle.Location += config.TagMargin;
				clientRectangle.Width -= config.TagMargin.Width * 2;
				clientRectangle.Height -= config.TagMargin.Height * 2;

				if (config.DecorationType == 1) { // 影
					int gap = GetShadowGap();
					Point pt = clientRectangle.Location + new Size(gap, gap);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
				} else if (config.DecorationType == 2) { // 縁取り
					int gap = GetHemGap();
					Point pt = clientRectangle.Location + new Size(0, -gap);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
					pt = clientRectangle.Location + new Size(-gap, 0);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
					pt = clientRectangle.Location + new Size(gap, 0);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
					pt = clientRectangle.Location + new Size(0, gap);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
					pt = clientRectangle.Location + new Size(-gap / 2, -gap / 2);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
					pt = clientRectangle.Location + new Size(gap / 2, -gap / 2);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
					pt = clientRectangle.Location + new Size(-gap / 2, gap / 2);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
					pt = clientRectangle.Location + new Size(gap / 2, gap / 2);
					g.DrawString(text, font, decoBrush, new Rectangle(pt, clientRectangle.Size), sf);
				}
				g.DrawString(text, font, fontBrush, clientRectangle, sf);
			}
		}

		private int GetShadowGap() { // 影の位置
			return font.Height / 32 + 1; // 適当に。。
		}
		private int GetHemGap() { // 縁取り位置
			return GetShadowGap() / 2 + 1; // 適当に。。
		}
	}
}
