﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;

namespace ふせん {
	public partial class TagForm : Form {
		TagData data;
		string tagsDir;

		string text;
		Bitmap bufferBmp;

		EditorForm editorForm = null;

		public delegate void EditDelegate(TagForm sender, EventArgs e);
		public event EditDelegate OnEdit;

		public TagData TagData { get { return data; } }
		public string FileName { get { return Path.Combine(tagsDir, data.FileName); } }

		public TagForm(TagData data, string tagsDir) {
			InitializeComponent();
			snapper1.Register(this);

			this.data = data;
			this.tagsDir = tagsDir;
			Location = data.Location;

			// ダミーを作成しておく
			bufferBmp = new Bitmap(1, 1);
		}

		/// <summary>
		/// 後始末
		/// </summary>
		private void TagForm_FormClosing(object sender, FormClosingEventArgs e) {
			data.Location = Location;
			bufferBmp.Dispose();
		}

		#region publicな操作

		/// <summary>
		/// ファイルの再読込
		/// </summary>
		/// <param name="redrawForce">中身変わってなくても再描画</param>
		public void ReloadTagFile(Config config, TagDrawer drawer, bool redrawForce) {
			string fileName = FileName;
			if (File.Exists(fileName)) {
				string newText = File.ReadAllText(fileName).Trim();
				if (redrawForce || text != newText) {
					// 変わってたら更新処理
					text = newText;
					using (Graphics g = Graphics.FromHwnd(Handle)) {
						Size size = drawer.GetSize(g, text);
						snapper1.ClientSize = size;
						bufferBmp.Dispose();
						bufferBmp = new Bitmap(size.Width, size.Height);
					}
					using (Graphics g = Graphics.FromImage(bufferBmp)) {
						drawer.Draw(g, text, ClientRectangle);
					}
					Invalidate();
					Update();
				}
			}
			// 透明度とかの反映
			Opacity = config.Opacity / 100.0;
			snapper1.Enable = config.Snap;
		}

		/// <summary>
		/// ファイル名変更時の処理
		/// </summary>
		public void OnChangeFileName(string newFileName) {
			data.FileName = newFileName;
		}

		/// <summary>
		/// 新しいタグが生成されたときに呼ばれる
		/// </summary>
		public void OnNewForm(TagForm form) {
			// お互いにSnapperに登録
			snapper1.Targets.Add(new Toolkit.Forms.SnapTargetControl(form));
			form.snapper1.Targets.Add(new Toolkit.Forms.SnapTargetControl(this));
		}

		/// <summary>
		/// タグが削除されるときに呼ばれる
		/// </summary>
		public void OnDeleteForm(TagForm form) {
			// 面倒なので自分側のsnapperからだけ削除
			snapper1.Targets.RemoveAll(delegate(Toolkit.Forms.ISnapTarget t) {
				Toolkit.Forms.SnapTargetControl c = t as Toolkit.Forms.SnapTargetControl;
				return c != null && c.Control == this;
			});
		}

		/// <summary>
		/// 編集処理
		/// </summary>
		public void Edit(Config config) {
			switch (config.EditorType) {
			case 0: // txtの関連付け
				{
					string tmpFile = Path.Combine(Path.GetDirectoryName(
						Process.GetCurrentProcess().MainModule.FileName), "tmp");
					while (File.Exists(tmpFile + ".txt")) {
						tmpFile += "$";
					}
					tmpFile += ".txt";
					string editor;
					try {
						File.WriteAllBytes(tmpFile, new byte[] { });
						editor = Toolkit.Windows.Assoc.FindExecutable(tmpFile);
					} finally {
						File.Delete(tmpFile);
					}

					if (File.Exists(editor)) {
						// コマンドラインは適当
						Process.Start(editor, "\"" + FileName + "\"");
					} else {
						editor = Environment.ExpandEnvironmentVariables("%windir%\\notepad.exe");
						if (File.Exists(editor)) {
							Process.Start(editor, "\"" + FileName + "\"");
						} else {
							Process.Start(FileName);
						}
					}
				}
				break;
			case 1: // 内蔵簡易エディタ
				if (editorForm == null || editorForm.IsDisposed) {
					editorForm = new EditorForm(this);
					editorForm.Show(this);
				} else {
					editorForm.Visible = false; // 念のため
					editorForm.Show(this);
					editorForm.Activate();
					editorForm.BringToFront();
				}
				break;
			case 2: // 指定エディタ
				try {
					// コマンドラインは適当
					Process.Start(config.EditorPath, "\"" + FileName + "\"");
				} catch (Win32Exception) {
					Process.Start(FileName);
				}
				break;
			}
		}

		/// <summary>
		/// 最前面表示にしたり最後面表示?にしたり
		/// </summary>
		public void SetTopMost(bool topMost) {
			TopMost = topMost;
			Activate();
			BringToFront();
		}

		#endregion

		/// <summary>
		/// 描画。
		/// </summary>
		private void TagForm_Paint(object sender, PaintEventArgs e) {
			e.Graphics.DrawImage(bufferBmp, e.ClipRectangle.X, e.ClipRectangle.Y, e.ClipRectangle, GraphicsUnit.Pixel);
		}

		enum SetWindowsPosPosition : int {
			HWND_TOP = 0,
			HWND_BOTTOM = 1,
			HWND_TOPMOST = -1,
			HWND_NOTOPMOST = -2
		}
		[System.Runtime.InteropServices.StructLayout(System.Runtime.InteropServices.LayoutKind.Sequential)]
		struct WINDOWPOS {
			public IntPtr hwnd;
			public IntPtr hwndInsertAfter;
			public int x;
			public int y;
			public int cx;
			public int cy;
			public uint flags;
		}
		/// <summary>
		/// NCHITTESTとか最背面表示とか
		/// </summary>
		protected override void WndProc(ref Message m) {
			// NCHITTESTしてるときのダブルクリック対策
			const int WM_SYSCOMMAND = 0x112;
			const int SC_MAXIMIZE = 0xF030;
			if (m.Msg == WM_SYSCOMMAND) {
				if ((m.WParam.ToInt32() & 0xFFF0) == SC_MAXIMIZE) {
					// ↓ダブルクリックとも限らん気もするが。。
					TagForm_DoubleClick(this, null);
					// 基本クラスには知らせずに戻る
					return;
				}
			}

			base.WndProc(ref m);

			if (!TopMost) { // 最前面表示でないなら最背面表示とする
				// 最背面表示
				const int WM_WINDOWPOSCHANGING = 0x46;
				if (m.Msg == WM_WINDOWPOSCHANGING) {
					WINDOWPOS pos = (WINDOWPOS)System.Runtime.InteropServices.Marshal.PtrToStructure(m.LParam, typeof(WINDOWPOS));
					pos.hwndInsertAfter = (IntPtr)SetWindowsPosPosition.HWND_BOTTOM;
					System.Runtime.InteropServices.Marshal.StructureToPtr(pos, m.LParam, false);
				}
			}

			// NCHITTEST
			const int WM_NCHITTEST = 0x84;
			const int HTCLIENT = 1;
			const int HTCAPTION = 2;
			if (m.Msg == WM_NCHITTEST && m.Result.ToInt32() == HTCLIENT) {
				if ((Form.MouseButtons & MouseButtons.Right) == 0) { // ←右クリックメニューな時は邪魔しないように。。
					// マウスがタイトルバーにあるふりをする
					m.Result = (IntPtr)HTCAPTION;
				}
			}
		}

		#region 右クリックメニューとか

		private void TagForm_DoubleClick(object sender, EventArgs e) {
			編集EToolStripMenuItem_Click(this, null);
		}

		private void 編集EToolStripMenuItem_Click(object sender, EventArgs e) {
			// 編集
			if (OnEdit != null) {
				OnEdit(this, null);
			}
		}

		private void 削除DToolStripMenuItem_Click(object sender, EventArgs e) {
			if (MessageBox.Show(this, FileName + " を削除しますか?", "確認",
				MessageBoxButtons.YesNo, MessageBoxIcon.Information) == DialogResult.Yes) {
				File.Delete(FileName);
			}
		}

		#endregion

	}
}